(function() {
  var JasmineHttpServerSpy, MockServer, _, bodyParser, debug, doneRejectHandler, doneResolveHandler, express, getRequestObject, q;

  express = require('express');

  q = require('q');

  _ = require('lodash');

  bodyParser = require('body-parser');

  debug = require('debug')('jasmine-http-spy');

  doneResolveHandler = function(done) {
    return function(result) {
      if (typeof done === "function") {
        done();
      }
      return result;
    };
  };

  doneRejectHandler = function(done) {
    return function(err) {
      if (done != null) {
        if (typeof done.fail === "function") {
          done.fail(err);
        }
      }
      throw err;
    };
  };

  getRequestObject = function(req) {
    return _.pick(req, 'params', 'query', 'body', 'headers', 'originalUrl');
  };

  MockServer = class MockServer {
    constructor(routes1, httpSpy) {
      this.routes = routes1;
      this.httpSpy = httpSpy;
    }

    setUpApplication() {
      var i, len, ref, results, route;
      this.app = express();
      this.app.use(bodyParser.json({
        type: ['application/json', 'application/*+json']
      }));
      this.app.use(bodyParser.urlencoded({
        extended: true
      }));
      this.app.use(function(req, res, next) {
        res.header("Access-Control-Allow-Origin", "*");
        res.header("Access-Control-Allow-Headers", "Origin, X-Requested-With, Content-Type, Accept");
        return next();
      });
      ref = this.routes;
      results = [];
      for (i = 0, len = ref.length; i < len; i++) {
        route = ref[i];
        debug('Registering mock endpoint', JSON.stringify(route));
        results.push(((route) => {
          return this.app[route.method](route.url, (req, res) => {
            var process, requestObject, resolveRequest, responseObject;
            process = this.httpSpy[route.handlerName];
            requestObject = getRequestObject(req);
            responseObject = process(requestObject);
            resolveRequest = function(responseObject) {
              var body, headers, statusCode;
              statusCode = responseObject.statusCode || 200;
              body = responseObject.body || {};
              headers = responseObject.headers || {};
              debug(`Responding to request: ${route.method} ${req.originalUrl}`);
              debug("Request: \n\t" + JSON.stringify(requestObject));
              debug("Response: \n\t" + JSON.stringify(responseObject));
              return res.status(statusCode).set(headers).send(body);
            };
            if (!responseObject) {
              return resolveRequest({
                statusCode: 500,
                body: "jasmine-http-server-spy was not given a response object -- this is probably a bug in your tests"
              });
            } else if (_.isFunction(responseObject.then)) {
              return responseObject.then(resolveRequest, function(reason) {
                return resolveRequest({
                  statusCode: 500,
                  body: "jasmine-http-server-spy caught rejected promise " + reason + " -- this is probably a bug in your tests"
                });
              });
            } else {
              return resolveRequest(responseObject);
            }
          });
        })(route));
      }
      return results;
    }

    start(...args) {
      var deferred, done;
      this.setUpApplication();
      if (args.length === 2 && typeof args[0] === 'number' && typeof args[1] === 'function') {
        this.server = this.app.listen(args[0]);
        done = args[1];
      } else {
        this.server = this.app.listen(...args);
        done = null;
      }
      deferred = q.defer();
      this.server.on('listening', deferred.resolve);
      this.server.on('error', deferred.reject);
      return deferred.promise.then(doneResolveHandler(done), doneRejectHandler(done));
    }

    stop(done) {
      return q.ninvoke(this.server, 'close').then(doneResolveHandler(done), doneRejectHandler(done));
    }

  };

  JasmineHttpServerSpy = (function() {
    var name, routes;

    class JasmineHttpServerSpy {
      constructor(_name, _routes) {
        routes = _routes;
        name = _name;
        if (!routes || routes.length === 0) {
          throw new Error("Routes list should be provided");
        }
        this.setUpSpies();
        this.server = new MockServer(routes, this);
      }

      setUpSpies() {
        var handlerName, handlerNames, i, len, results;
        handlerNames = _.map(routes, 'handlerName');
// Default implementation to 404
        results = [];
        for (i = 0, len = handlerNames.length; i < len; i++) {
          handlerName = handlerNames[i];
          this[handlerName] = jasmine.createSpy(`${name}.${handlerName}`);
          results.push(this[handlerName].and.returnValue({
            statusCode: 404,
            body: {
              message: 'Page not found'
            }
          }));
        }
        return results;
      }

    };

    routes = null;

    name = null;

    return JasmineHttpServerSpy;

  }).call(this);

  module.exports = {
    createSpyObj: function(name, routes) {
      return new JasmineHttpServerSpy(name, routes);
    }
  };

}).call(this);
